from .intermediate_statistics import accumulated_intermediate_statistics, intermediate_statistics_deltas
from .precision_recall import precision_recall_curve, fscore_curve, fscore, bootstrapped_fscore, best_fscore
from .error_rate import error_rate_curve, error_rate, best_error_rate
from . import reference



from sed_scores_eval.base_modules.error_rate import (
    single_error_rate_from_intermediate_statistics,
    best_error_rate_from_intermediate_statistics,
    error_rate_curve_from_intermediate_statistics
)
from sed_scores_eval.collar_based.intermediate_statistics import accumulated_intermediate_statistics


def error_rate_curve(
        scores, ground_truth, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0.,
        time_decimals=6, num_jobs=1,
):
    """Compute collar-based error rates with corresponding insertion rate,
    deletion rate and intermediate statistics for various operating points

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns: (all arrays sorted by corresponding score)
        error_rates ((dict of) 1d np.ndarray): error rates for all operating points
        insertion_rate ((dict of) 1d np.ndarray): insertion rates for all operating points
        deletion_rates ((dict of) 1d np.ndarray): deletion rates for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding error rate.
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return error_rate_curve_from_intermediate_statistics(intermediate_stats)


def error_rate(
        scores, ground_truth, threshold, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0.,
        return_onset_offset_dist_sum=False, time_decimals=6, num_jobs=1,
):
    """Compute collar-based error rates with corresponding insertion rate,
    deletion rate and intermediate statistics intermediate statistics for a
    specific decision threshold

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        threshold ((dict of) float): threshold that is to be evaluated.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        error_rate ((dict of) float): error rate for threshold
        insertion_rate ((dict of) float): insertion rate for threshold
        deletion rate ((dict of) float): deletion rate for threshold
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count for threshold
            'fps' (int): false positive count for threshold
            'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        return_onset_offset_dist_sum=return_onset_offset_dist_sum,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return single_error_rate_from_intermediate_statistics(
        intermediate_stats, threshold=threshold,
    )


def best_error_rate(
        scores, ground_truth, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0.,
        max_insertion_rate=None, max_deletion_rate=None,
        time_decimals=6, num_jobs=1,
):
    """Get the best possible (macro-averaged) collar-based error rate with
    corresponding insertion rate, deletion rate, intermediate statistics and
    decision threshold

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        max_insertion_rate: a maximal insertion rate that must not be exceeded.
        max_deletion_rate: a maximal deletion rate that must not be exceeded.
            If the constraint(s) cannot be achieved at any threshold, however,
            er, ir, dr and threshold of 1,0,1,inf are returned.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        error_rate ((dict of) float): best achievable error_rate value
        insertion_rate ((dict of) float): insertion rate at best error rate
        deletion_rate ((dict of) float): deletion rate at best error rate
        threshold ((dict of) float): threshold to obtain best error rate which
            is centered between the score that the threshold has to fall below
            and the next smaller score which results in different intermediate
            statistics.
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count at best error rate
            'fps' (int): false positive count at best error rate
            'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return best_error_rate_from_intermediate_statistics(
        intermediate_stats, max_insertion_rate=max_insertion_rate,
        max_deletion_rate=max_deletion_rate,
    )



from pathlib import Path
from . import base_modules
from .base_modules import io
from . import intersection_based
from . import collar_based
from . import segment_based
from . import clip_based
from . import utils

package_dir = Path(__file__).parent.parent.absolute()



import numpy as np
import pandas as pd
from sed_scores_eval.base_modules.scores import validate_score_dataframe


def onset_offset_curves(scores, timestamps, change_point_candidates=None):
    """get onset and offset times of event detections for various decision
    thresholds. Here, the number of event detections is given by the number of
    local maxima in the score signal with events being spawned when the
    decision threshold falls below the local maximum. However, usually only a
    subset of these events is active simultanously while others are inactive,
    e.g., because a certain threshold does not yet fall below all local
    maximums. For inactive events we return offset_time = onset_time. Further,
    when the decision threshold falls below a local minimum, two separate
    events merge into a single event. In this case, we keep the earlier event
    active with corresponding onset and offset times, while the later event is
    set inactive with offset_time = onset_time.

    ! There is a faster implementation of this function in cy_detection.pyx !

    Args:
        scores (1d np.ndarray): SED scores for a single event class
        timestamps (1d np.ndarray): onset timestamps for each score plus one more
            timestamp which is the final offset time.
        change_point_candidates (1d np.ndarray)

    Returns:
        scores_unique (1d np.ndarray): unique and sorted score array
        onset_times (2d np.ndarray): onset times for each possible event at
            each decsion threshold that falls below one of the unique scores.
            Shape is (number of unique scores, number of events/local maximums).
        offset_times (2d np.ndarray): offset times for each possible event at
            each decsion threshold that falls below one of the unique scores.
            Shape is (number of unique scores, number of events/local maximums).

    >>> y = np.array([.4,1.,.6,.6,1.,1.,.4])
    >>> ts = np.linspace(0., len(y)*.2, len(y) + 1)  # each score has width of 200ms
    >>> y, t_on, t_off = onset_offset_curves(y, ts)
    >>> y
    array([0.4, 0.6, 1. ])
    >>> np.stack((t_on, t_off), axis=-1)
    array([[[0. , 1.4],
            [0. , 0. ]],
    <BLANKLINE>
           [[0.2, 1.2],
            [0. , 0. ]],
    <BLANKLINE>
           [[0.2, 0.4],
            [0.8, 1.2]]])
    """
    scores = np.asanyarray(scores)
    if not scores.ndim == 1:
        raise ValueError(
            f'scores must be 1-dimensional array of single class SED scores, '
            f'but array of shape {scores.shape} was given.'
        )
    timestamps = np.asanyarray(timestamps)
    if not timestamps.ndim == 1 or len(timestamps) != (len(scores) + 1):
        raise ValueError(
            f'timestamps must be 1-dimensional array of length(len(scores) + 1), '
            f'but array of shape {timestamps.shape} was given.'
        )

    if change_point_candidates is None:
        change_point_candidates = np.unique(scores)
    detection = scores >= change_point_candidates[:, None]
    prev_detection = np.concatenate(
        (np.zeros_like(detection[:, :1]), detection),
        axis=1,
    )
    detection = np.concatenate(
        (detection, np.zeros_like(detection[:, :1]),),
        axis=1,
    )
    onsets = detection > prev_detection
    offsets = detection < prev_detection
    # assert (onsets.sum(-1) == offsets.sum(-1)).all()
    n_events = onsets.sum(-1)
    max_events = n_events.max()

    onset_times = np.zeros((len(change_point_candidates), max_events))
    offset_times = np.zeros_like(onset_times)
    thres_indices, frame_indices = np.argwhere(onsets).T
    n_events_offset = np.cumsum(n_events) - n_events
    event_indices = np.arange(len(thres_indices)) - n_events_offset[thres_indices]
    onset_times[thres_indices, event_indices] = timestamps[frame_indices]
    thres_indices, frame_indices = np.argwhere(offsets).T
    offset_times[thres_indices, event_indices] = timestamps[frame_indices]
    # print(onset_times.shape)
    return change_point_candidates, onset_times, offset_times


def onset_deltas(scores):
    """return the change in the total number of onsets when decision threshold
    falls below each of the scores, i.e., +1 at local maximums and -1 at local
    minimums in score signal.

    Args:
        scores (1d np.ndarray): SED scores for a single event class

    Returns:
        onset_deltas (1d np.ndarray): array with same length as scores
        indicating the change in the number of onsets when decision threshold
        falls below each of the scores, i.e., +1 at local maxima and -1 at
        local minima in score signal.

    >>> onset_deltas(np.array([1,2,3,3,4,3]))
    """
    assert isinstance(scores, np.ndarray), scores
    prev_scores = np.concatenate(([-np.inf], scores[:-1]))
    next_scores = np.concatenate((scores[1:], [-np.inf]))
    return (
        (scores > prev_scores).astype(int)
        - (next_scores > scores).astype(int)
    ) * (scores > -np.inf)


def scores_to_event_list(scores, thresholds, event_classes=None):
    """detect events and return as list

    Args:
        scores ((dict of) pandas.DataFrame): containing onset and offset times
            of a score window in first two columns followed by sed score
            columns for each event class.
        thresholds (1d np.ndarray or dict of floats): thresholds to be used for
            the different event classes.
        event_classes (list of str): optional list of event classes used to
            assert correct event labels in scores DataFrame

    Returns:
        event_list (list of tuple): list of events as tuples (onset, offset, event_class)
    """
    if not isinstance(scores, pd.DataFrame) and hasattr(scores, 'keys'):
        assert callable(scores.keys)
        keys = sorted(scores.keys())
        _, event_classes = validate_score_dataframe(
            scores[keys[0]], event_classes=event_classes)
        if isinstance(thresholds, dict):
            thresholds = np.array([
                thresholds[event_class] for event_class in event_classes])
        return {
            key: scores_to_event_list(
                scores[key], thresholds, event_classes=event_classes)
            for key in keys
        }
    timestamps, event_classes = validate_score_dataframe(
        scores, event_classes=event_classes)
    onset_times = scores['onset'].to_numpy()
    offset_times = scores['offset'].to_numpy()
    scores = scores[event_classes].to_numpy()
    if isinstance(thresholds, dict):
        thresholds = np.array([
            thresholds[event_class] for event_class in event_classes])
    detections = scores > thresholds
    zeros = np.zeros_like(detections[:1, :])
    detections = np.concatenate((zeros, detections, zeros), axis=0).astype(float)
    change_points = detections[1:] - detections[:-1]
    event_list = []
    for k in np.argwhere(np.abs(change_points).max(0) > .5).flatten():
        onsets = np.argwhere(change_points[:, k] > .5).flatten()
        offsets = np.argwhere(change_points[:, k] < -.5).flatten()
        assert len(onsets) == len(offsets) > 0
        for onset, offset in zip(onsets, offsets):
            event_list.append((
                onset_times[onset], offset_times[offset-1],
                event_classes[k]
            ))
    return sorted(event_list)



from .intermediate_statistics import accumulated_intermediate_statistics
from .precision_recall import (
    precision_recall_curve,
    fscore_curve, fscore, best_fscore,
    average_precision,
)
from .error_rate import error_rate_curve, error_rate, best_error_rate
from .roc import roc_curve, auroc
from . import reference



import numpy as np
from numbers import Number
from sed_scores_eval.utils.array_ops import get_first_index_where


def multi_label_to_single_label_ground_truths(ground_truth, event_classes):
    """convert dict of multi label ground truths to list of dict of single
    label ground truths.

    Args:
        ground_truth (dict of lists of tuples): list of ground truth event
            tuples (onset, offset, event label) for each audio clip.
        event_classes (list of str): list of event classes.

    Returns (dict of dicts of lists of tuple): list of single class ground
        truth event tuples (onset, offset) for each audio clip for each event
        class

    """
    if not isinstance(ground_truth, dict):
        raise ValueError(
            f'ground_truth must be of type dict but {type(ground_truth)} was '
            f'given.'
        )
    single_label_ground_truths = {
        class_name: {} for class_name in event_classes}
    n_ref_count = {
        class_name: 0 for class_name in event_classes}
    for clip_id in ground_truth.keys():
        for class_name in event_classes:
            single_label_ground_truths[class_name][clip_id] = []
        if not isinstance(ground_truth[clip_id], (list, tuple)):
            raise ValueError(
                f'ground_truth values must be of type list/tuple but '
                f'{type(ground_truth[clip_id])} was found for key {clip_id}.'
            )
        for i in range(len(ground_truth[clip_id])):
            if (
                not isinstance(ground_truth[clip_id][i], (list, tuple))
                or len(ground_truth[clip_id][i]) != 3
            ):
                raise ValueError(
                    f'ground_truth event tuples must have the form '
                    f'(onset_time, offset_time, event label) but '
                    f'{ground_truth[clip_id][i]} was given for key {clip_id}.'
                )
            t_on, t_off, event_label = ground_truth[clip_id][i]
            if (
                not isinstance(t_on, Number) or not isinstance(t_off, Number)
                or not isinstance(event_label, (str, int))
            ):
                raise ValueError(
                    f'ground_truth event tuples must have the form '
                    f'(onset time, offset time, event label) with onset time '
                    f'and offset time being numbers and event label either '
                    f'being integer or string but types {type(t_on)}, '
                    f'{type(t_off)} and {type(event_label)} were given for '
                    f'key {clip_id}.'
                )
            if event_label not in event_classes:
                raise ValueError(
                    f'event label {event_label} for key {clip_id} is not listed '
                    f'in event_classes {event_classes}.'
                )
            single_label_ground_truths[event_label][clip_id].append((t_on, t_off))
            n_ref_count[event_label] += 1
    classes_without_ground_truth_events = [
        class_name for class_name, n_ref in n_ref_count.items() if n_ref == 0
    ]
    assert len(classes_without_ground_truth_events) == 0, f"No ground truth events for classes {classes_without_ground_truth_events}."
    return single_label_ground_truths


def event_counts_and_durations(ground_truth, event_classes):
    """obtain the total number and combined duration of ground truths events
    for each event class.

    Args:
        ground_truth (dict of lists of tuples): list of ground truth event
            tuples (onset, offset, event label) for each audio clip.
        event_classes (list of str): list of event classes.

    Returns:
        counts (dict of int): total number of ground truth events for each
            event class
        durations (dict of float): combined duration of ground truth events for
            each event class

    """
    if not isinstance(ground_truth, dict):
        raise ValueError(
            f'ground_truth must be of type dict but {type(ground_truth)} was'
            f'given.'
        )
    counts = {key: 0 for key in event_classes}
    durations = {key: 0. for key in event_classes}
    for key in ground_truth.keys():
        if not isinstance(ground_truth[key], (list, tuple)):
            raise ValueError(
                f'ground_truth values must be of type list/tuple but '
                f'{type(ground_truth[key])} was found for key {key}.'
            )
        for i in range(len(ground_truth[key])):
            if (
                not isinstance(ground_truth[key][i], (list, tuple))
                or len(ground_truth[key][i]) != 3
            ):
                raise ValueError(
                    f'ground_truth event tuples must have the form '
                    f'(onset_time, offset_time, event label) but '
                    f'{ground_truth[key][i]} was given for key {key}.'
                )
            t_on, t_off, event_label = ground_truth[key][i]
            if (
                not isinstance(t_on, Number) or not isinstance(t_off, Number)
                or not isinstance(event_label, (str, int))
            ):
                raise ValueError(
                    f'ground_truth event tuples must have the form '
                    f'(onset time, offset time, event label) with onset time '
                    f'and offset time being numbers and event label either '
                    f'being integer or string but types {type(t_on)}, '
                    f'{type(t_off)} and {type(event_label)} were given for '
                    f'key {key}.'
                )
            if event_label not in event_classes:
                raise ValueError(
                    f'event label {event_label} for key {key} is not listed '
                    f'in event_classes {event_classes}.'
                )
            counts[event_label] += 1
            durations[event_label] += t_off - t_on
    return counts, durations


def onset_offset_times_to_indices(onset_time, offset_time, timestamps):
    """convert an onset/offset time pair to the indices of the frames in which
    the onset/offset time lie (or the last frame index if the offset_time lies
    beyond timestamps[-1]).

    Args:
        onset_time (float):
        offset_time (float):
        timestamps (1d np.array): onset timestamps for each frame plus one more
            timestamp which is the final offset time.

    Returns:
        onset_idx:
        offset_idx:

    """
    if not np.isscalar(onset_time) or not np.isscalar(offset_time):
        raise ValueError(
            f'onset_time and offset_time must be scalars, but '
            f'{type(onset_time)} and {type(offset_time)} were given.'
        )
    timestamps = np.asanyarray(timestamps)
    if not timestamps.ndim == 1:
        raise ValueError(
            f'timestamps must be 1-dimensional array but array of shape '
            f'{timestamps.shape} was given.'
        )
    if not offset_time > onset_time >= timestamps[0] >= 0.:
        raise ValueError(
            f'offset_time must be greater than onset_time which must be '
            f'greater equal timestamps[0] which must be greater equal 0. '
            f'However, offset time is {offset_time}, onset time is '
            f'{onset_time} and timestamps are {timestamps}.'
        )
    onset_idx = max(
        get_first_index_where(timestamps, 'gt', onset_time) - 1,
        0
    )
    offset_idx = min(
        get_first_index_where(timestamps, 'geq', offset_time),
        len(timestamps) - 1,
    )
    return onset_idx, offset_idx



import numpy as np
from sed_scores_eval.base_modules.curves import xsort, staircase_auc


def roc_curve_from_intermediate_statistics(scores_intermediate_statistics):
    """compute ROC (TPR-FPR) curve from intermediate statistics

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.

    Returns: (all arrays sorted by corresponding false positive rate)
        true_positive_rates ((dict of) 1d np.ndarray): TPR values for all operating points
        false_positive_rates ((dict of) 1d np.ndarray): FPR values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding TPR-FPR pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events

    """
    if isinstance(scores_intermediate_statistics, dict):
        roc_curves = {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            roc_curves[class_name] = roc_curve_from_intermediate_statistics(scores_stats)
        return roc_curves

    if not isinstance(scores_intermediate_statistics, (list, tuple)):
        raise ValueError(
            f'scores_intermediate_statistics must be list/tuple of length 2, '
            f'but {type(scores_intermediate_statistics)} was given.'
        )
    if len(scores_intermediate_statistics) != 2:
        raise ValueError(
            f'scores_intermediate_statistics must be list/tuple of length 2, '
            f'but list/tuple of length {len(scores_intermediate_statistics)}'
            f'was given.'
        )
    scores, stats = scores_intermediate_statistics
    if not isinstance(stats, dict):
        raise ValueError(
            f'intermediate_statistics must be dict but {type(stats)} was given.')
    if not all([key in stats for key in ['tps', 'fps', 'n_ref']]):
        raise ValueError(
            f'intermediate_statistics must contain keys "tps", "fps" and "n_ref". '
            f'Provided keys: {sorted(stats.keys())}.'
        )

    tpr = stats['tps'] / stats['n_ref']
    fpr = stats['fps'] / (stats['tns'] + stats['fps'])

    return xsort(tpr, fpr, scores, stats)


def auroc_from_intermediate_statistics(scores_intermediate_statistics, max_fpr=None):
    """compute area under ROC curve

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.
        max_fpr (float): maximum false positive rate up to which to compute partial auc

    Returns:
        auroc ((dict of) float): mean and class-wise area under ROC curve
        roc_curve: ROC (TPR-FPR) curve(s) as provided by
            `roc_curve_from_intermediate_statistics`

    """
    if isinstance(scores_intermediate_statistics, dict):
        auroc, roc_curves = {}, {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            auroc[class_name], roc_curves[class_name] = auroc_from_intermediate_statistics(
                scores_stats, max_fpr=max_fpr,
            )
        auroc['mean'] = np.mean([auroc[class_name] for class_name in auroc])
        return auroc, roc_curves
    roc_curve = roc_curve_from_intermediate_statistics(
        scores_intermediate_statistics
    )
    tpr, fpr, *_ = roc_curve
    norm = 1 if max_fpr is None else max_fpr
    auroc = staircase_auc(tpr, fpr, max_x=max_fpr)/norm
    return auroc, roc_curve



import numpy as np
from scipy.interpolate import interp1d


def assert_roc_geq_roc_ref(roc, roc_ref, class_name='', upper_bound=None):
    tpr, fpr, *_ = roc
    fpr = np.round(fpr, 6)
    tpr_ref, fpr_ref, *_ = roc_ref
    fpr_ref = np.round(fpr_ref, 6)
    tpr = interp1d(
        fpr, tpr, kind='previous',
        bounds_error=False, fill_value=(0, tpr[-1])
    )(fpr_ref)
    assert (tpr >= tpr_ref).all(), (
        class_name,
        np.sum(tpr < tpr_ref),
        len(tpr),
        (tpr_ref - tpr).max(),
    )
    if upper_bound is not None:
        assert ((tpr - tpr_ref) < upper_bound).all(), (
            class_name,
            np.sum((tpr - tpr_ref) >= upper_bound),
            len(tpr),
            (tpr - tpr_ref).max(),
        )



import numpy as np
from sed_scores_eval.base_modules.curves import get_curve_idx_for_threshold, xsort


def precision_recall_curve_from_intermediate_statistics(
        scores_intermediate_statistics
):
    """Compute precision-recall curve from intermediate_statistics

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided, keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.

    Returns: (all arrays sorted by corresponding recall)
        precisions ((dict of) 1d np.ndarray): precision values for all operating points
        recalls ((dict of) 1d np.ndarray): recall values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding precision-recall pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events

    """
    if isinstance(scores_intermediate_statistics, dict):
        pr_curves = {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            pr_curves[class_name] = precision_recall_curve_from_intermediate_statistics(scores_stats)
        return pr_curves

    if not isinstance(scores_intermediate_statistics, (list, tuple)):
        raise ValueError(
            f'scores_intermediate_statistics must be list/tuple of length 2, '
            f'but {type(scores_intermediate_statistics)} was given.'
        )
    if len(scores_intermediate_statistics) != 2:
        raise ValueError(
            f'scores_intermediate_statistics must be list/tuple of length 2, '
            f'but list/tuple of length {len(scores_intermediate_statistics)}'
            f'was given.'
        )
    scores, stats = scores_intermediate_statistics
    if not isinstance(stats, dict):
        raise ValueError(
            f'intermediate_statistics must be dict but {type(stats)} was given.')
    if not all([key in stats for key in ['tps', 'fps', 'n_ref']]):
        raise ValueError(
            f'intermediate_statistics must contain keys "tps", "fps" and "n_ref". '
            f'Provided keys: {sorted(stats.keys())}.'
        )
    p = stats['tps'] / np.maximum(stats['tps']+stats['fps'], 1)
    p[(stats['tps']+stats['fps']) == 0] = 1.
    if stats['n_ref'] == 0:
        raise ValueError('Recall not defined if the are no positive instances')
    r = stats['tps'] / stats['n_ref']

    return xsort(p, r, scores, stats)


def fscore_from_precision_recall(precision, recall, *, beta=1.):
    """Compute f-score from precision and recall.

    Args:
        precision (scalar or np.ndarray):
        recall (scalar or np.ndarray):
        beta: \beta parameter for f-score computation

    Returns: f-score

    """
    return (
        (1 + beta**2) * precision * recall
        / np.maximum(beta**2 * precision + recall, 1e-15)
    )


def fscore_curve_from_intermediate_statistics(
        scores_intermediate_statistics, beta=1.
):
    """Compute f-scores with corresponding precisions, recalls and
    intermediate statistics for various operating points

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.
        beta: \beta parameter of f-score computation

    Returns: (all arrays sorted by corresponding score)
        f_beta ((dict of) 1d np.ndarray): f-score values  for all operating
            points
        precisions ((dict of) 1d np.ndarray): precision values for all operating points
        recalls ((dict of) 1d np.ndarray): recall values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding precision-recall pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps': 1d np.ndarray of true positive counts for each score
            'fps': 1d np.ndarray of false positive counts for each score
            'n_ref': integer number of ground truth events

    """
    if isinstance(scores_intermediate_statistics, dict):
        fscore_curves = {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            fscore_curves[class_name] = fscore_curve_from_intermediate_statistics(
                scores_stats, beta=beta,
            )
        return fscore_curves
    p, r, scores, intermediate_stats = precision_recall_curve_from_intermediate_statistics(
        scores_intermediate_statistics
    )
    f_beta = fscore_from_precision_recall(p, r, beta=beta)
    f_beta, scores, p, r, intermediate_stats = xsort(f_beta, scores, p, r, intermediate_stats)
    return f_beta, p, r, scores, intermediate_stats


def best_fscore_from_intermediate_statistics(
        scores_intermediate_statistics,
        beta=1., min_precision=0., min_recall=0.
):
    """Get the best possible (macro-average) f-score with corresponding
    precision, recall, intermediate statistics and decision threshold

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.
        beta: \beta parameter of f-score computation
        min_precision: the minimum precision that must be achieved.
        min_recall: the minimum recall that must be achieved. If the
            constraint(s) cannot be achieved at any threshold, however,
            fscore, precision, recall and threshold of 0,1,0,inf are returned.

    Returns:
        f_beta ((dict of) float): best achievable f-score value
        precision ((dict of) float): precision value at best fscore
        recall ((dict of) float): recall value at best fscore
        threshold ((dict of) float): threshold to obtain best fscore which is
            centered between the score that the threshold has to fall below
            and the next smaller score which results in different intermediate
            statistics.
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count at best fscore
            'fps' (int): false positive count at best fscore
            'n_ref' (int): number of ground truth events

    """
    if isinstance(scores_intermediate_statistics, dict):
        f, p, r, thresholds, intermediate_stats = {}, {}, {}, {}, {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            (
                f[class_name], p[class_name], r[class_name],
                thresholds[class_name], intermediate_stats[class_name]
            ) = best_fscore_from_intermediate_statistics(
                scores_stats, beta=beta,
                min_precision=min_precision,
                min_recall=min_recall,
            )
        f['macro_average'] = np.mean([f[class_name] for class_name in f])
        p['macro_average'] = np.mean([p[class_name] for class_name in p])
        r['macro_average'] = np.mean([r[class_name] for class_name in r])
        (
            f['micro_average'], p['micro_average'], r['micro_average']
        ) = micro_average(intermediate_stats, beta=beta)
        return f, p, r, thresholds, intermediate_stats
    f, p, r, scores, intermediate_stats = \
        fscore_curve_from_intermediate_statistics(
            scores_intermediate_statistics, beta=beta
        )
    f[p < min_precision] = 0.
    f[r < min_recall] = 0.
    best_idx = len(f) - 1 - np.argmax(f[::-1], axis=0)
    threshold = (
        (scores[best_idx] + scores[best_idx-1])/2 if best_idx > 0 else -np.inf
    )

    return (
        f[best_idx], p[best_idx], r[best_idx], threshold,
        _recursive_get_item(intermediate_stats, best_idx)
    )


def single_fscore_from_intermediate_statistics(
        scores_intermediate_statistics, threshold, *, beta=1., ):
    """Get f-score with corresponding precision, recall and intermediate
    statistics for a specific decision threshold

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.
        threshold ((dict of) float): threshold that is to be evaluated.
        beta: \beta parameter of f-score computation

    Returns:
        fscore ((dict of) float): fscore value for threshold
        precision ((dict of) float): precision value for threshold
        recall ((dict of) float): recall value for threshold
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count for threshold
            'fps' (int): false positive count for threshold
            'n_ref' (int): number of ground truth events

    """
    if isinstance(threshold, dict):
        if not all([np.isscalar(thr) for thr in threshold]):
            raise ValueError('All values of thresholds dict must be scalars')
    elif not np.isscalar(threshold):
        raise ValueError(
            f'threshold must be (dict of) scalar(s) but '
            f'{type(threshold)} was given.'
        )
    if isinstance(scores_intermediate_statistics, dict):
        f, p, r, intermediate_stats = {}, {}, {}, {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            (
                f[class_name], p[class_name], r[class_name],
                intermediate_stats[class_name]
            ) = single_fscore_from_intermediate_statistics(
                scores_stats,
                threshold[class_name] if isinstance(threshold, dict)
                else threshold,
                beta=beta,
            )
        f['macro_average'] = np.mean([f[class_name] for class_name in f])
        p['macro_average'] = np.mean([p[class_name] for class_name in p])
        r['macro_average'] = np.mean([r[class_name] for class_name in r])
        (
            f['micro_average'], p['micro_average'], r['micro_average']
        ) = micro_average(intermediate_stats, beta=beta)
        return f, p, r, intermediate_stats
    pr_curve = precision_recall_curve_from_intermediate_statistics(
        scores_intermediate_statistics
    )
    p, r, scores, intermediate_stats = pr_curve
    return _single_fscore_from_precision_recall_curve(
        p, r, scores, threshold, beta=beta,
        intermediate_statistics=intermediate_stats
    )


def _single_fscore_from_precision_recall_curve(
        precisions, recalls, scores, threshold, beta=1.,
        intermediate_statistics=None
):
    idx = get_curve_idx_for_threshold(scores, threshold)
    p = precisions[idx]
    r = recalls[idx]
    f = fscore_from_precision_recall(p, r, beta=beta)
    if intermediate_statistics is None:
        return f, p, r
    else:
        return f, p, r, _recursive_get_item(intermediate_statistics, idx)


def micro_average(intermediate_stats, beta=1.):
    """Compute the mirco averaged f-score, where the intermediate statistics
    are summed up before computation of precision, recall and f-score

    Args:
        intermediate_stats (dict of dict): contains a dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count for threshold
            'fps' (int): false positive count for threshold
            'n_ref' (int): number of ground truth events
            for each event class
        beta: \beta parameter of f-score computation

    Returns:
        fscore (float): micro-average fscore
        precision (float): micro-average precision
        recall (float): micro-average recall

    """
    tps = np.sum([
        intermediate_stats[class_name]['tps']
        for class_name in intermediate_stats
    ])
    fps = np.sum([
        intermediate_stats[class_name]['fps']
        for class_name in intermediate_stats
    ])
    n_ref = np.sum([
        intermediate_stats[class_name]['n_ref']
        for class_name in intermediate_stats
    ])
    p = tps / np.maximum(tps + fps, 1)
    r = tps / np.maximum(n_ref, 1)
    f = fscore_from_precision_recall(p, r, beta=beta)
    return f, p, r


def fscore_from_sed_eval_metrics(sed_eval_metrics):
    """extract class-wise and averaged fscores, precisions and recalls from
    sed_eval metrics object

    Args:
        sed_eval_metrics:

    Returns:
        fscore (dict of float): fscore values
        precision (dict of float): precision values
        recall (dict of float): recall values

    """
    f = {}
    p = {}
    r = {}
    sed_eval_results_classwise = sed_eval_metrics.results_class_wise_metrics()
    for key in sed_eval_results_classwise:
        f[key] = sed_eval_results_classwise[key]['f_measure']['f_measure']
        p[key] = sed_eval_results_classwise[key]['f_measure']['precision']
        r[key] = sed_eval_results_classwise[key]['f_measure']['recall']
    sed_eval_results_macro = sed_eval_metrics.results_class_wise_average_metrics()
    f['macro_average'] = sed_eval_results_macro['f_measure']['f_measure']
    p['macro_average'] = sed_eval_results_macro['f_measure']['precision']
    r['macro_average'] = sed_eval_results_macro['f_measure']['recall']
    sed_eval_results_micro = sed_eval_metrics.results_overall_metrics()
    f['micro_average'] = sed_eval_results_micro['f_measure']['f_measure']
    p['micro_average'] = sed_eval_results_micro['f_measure']['precision']
    r['micro_average'] = sed_eval_results_micro['f_measure']['recall']
    return f, p, r


def average_precision_from_intermediate_statistics(
        scores_intermediate_statistics):
    """Compute the average precision

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.

    Returns:
        average_precision ((dict of) float): mean and class-wise average
            precisions
        pr_curve: Precision-Recall curve(s) as provided by
            `precision_recall_curve_from_intermediate_statistics`
    """
    if isinstance(scores_intermediate_statistics, dict):
        ap, pr_curves = {}, {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            ap[class_name], pr_curves[class_name] = average_precision_from_intermediate_statistics(
                scores_stats,
            )
        ap['mean'] = np.mean([ap[class_name] for class_name in ap])
        return ap, pr_curves

    pr_curve = precision_recall_curve_from_intermediate_statistics(
        scores_intermediate_statistics
    )
    p, r, scores, intermediate_stats = pr_curve
    return average_precision_from_precision_recall_curve(p, r), pr_curve


def average_precision_from_precision_recall_curve(p, r):
    """compute average precision from Precision-Recall curve

    Args:
        p: Precision for each operating point
        r: Recall for each operating point

    Returns:

    """
    _, unique_recall_indices = np.unique(r[::-1], return_index=True)
    unique_recall_indices = - 1 - unique_recall_indices
    r = r[unique_recall_indices]
    p = p[unique_recall_indices]
    return np.sum(p[1:] * (r[1:]-r[:-1]))


def _recursive_get_item(stats, idx):
    if isinstance(stats, dict):
        return {key: _recursive_get_item(stats[key], idx) for key in stats}
    if np.isscalar(stats):
        return stats
    return stats[idx]



import numpy as np
import multiprocessing
from sed_scores_eval.utils.array_ops import get_first_index_where
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from sed_scores_eval.base_modules import statistics
from sed_scores_eval.base_modules.io import parse_inputs, parse_audio_durations
from sed_scores_eval.base_modules.ground_truth import multi_label_to_single_label_ground_truths


def intermediate_statistics_deltas(
        scores, ground_truth, audio_durations, *,
        segment_length=1., time_decimals=6, num_jobs=1,
):
    if not isinstance(num_jobs, int) or num_jobs < 1:
        raise ValueError(
            f'num_jobs has to be an integer greater or equal to 1 but '
            f'{num_jobs} was given.'
        )
    scores, ground_truth, audio_ids = parse_inputs(scores, ground_truth)
    if audio_durations is not None:
        audio_durations = parse_audio_durations(audio_durations, audio_ids=audio_ids)

    _, event_classes = validate_score_dataframe(scores[audio_ids[0]])
    single_label_ground_truths = multi_label_to_single_label_ground_truths(
        ground_truth, event_classes)

    if num_jobs == 1:
        deltas = _worker(
            audio_ids, scores, single_label_ground_truths, audio_durations,
            segment_length,  event_classes, time_decimals,
        )
    else:
        deltas = {}
        queue = multiprocessing.Queue()
        shard_size = int(np.ceil(len(audio_ids) / num_jobs))
        shards = [
            audio_ids[i*shard_size:(i+1)*shard_size] for i in range(num_jobs)
            if i*shard_size < len(audio_ids)
        ]
        processes = [
            multiprocessing.Process(
                target=_worker,
                args=(
                    shard, scores, single_label_ground_truths, audio_durations,
                    segment_length,  event_classes, time_decimals, queue
                ),
                daemon=True,
            )
            for shard in shards
        ]
        try:
            for p in processes:
                p.start()
            count = 0
            while count < len(shards):
                deltas_i = queue.get()
                assert len(deltas.keys() & deltas_i.keys()) == 0, sorted(deltas.keys() & deltas_i.keys())
                deltas.update(deltas_i)
                count += 1
        finally:
            for p in processes:
                p.terminate()
    return deltas


def accumulated_intermediate_statistics_from_deltas(deltas):
    multi_label_statistics = statistics.accumulated_intermediate_statistics_from_deltas(deltas)
    for _, stats_cls in multi_label_statistics.values():
        stats_cls['n_sys'] = stats_cls['tps'] + stats_cls['fps']
        stats_cls['n_ref'] = stats_cls['tps'][0]
        fns = stats_cls['n_ref'] - stats_cls['tps']
        stats_cls['tns'] = stats_cls['n_sys'][0] - stats_cls['n_sys'] - fns
    return multi_label_statistics


def accumulated_intermediate_statistics(
        scores, ground_truth, audio_durations, *, deltas=None,
        segment_length=1., time_decimals=6, num_jobs=1,
):
    """Cascade of `intermediate_statistics_deltas` and `accumulated_intermediate_statistics_from_deltas`.

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        segment_length: the segment length of the segments that are to be
            evaluated.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high detected or ground truth events that have
            onsets or offsets right on a segment boundary may swap over to the
            adjacent segment because of small deviations due to limited
            floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:

    """
    if deltas is None:
        scores, ground_truth, audio_ids = parse_inputs(scores, ground_truth)
        deltas = intermediate_statistics_deltas(
            scores=scores, ground_truth=ground_truth,
            audio_durations=audio_durations,
            segment_length=segment_length, time_decimals=time_decimals,
            num_jobs=num_jobs,
        )
    else:
        audio_ids = list(deltas.keys())
    return accumulated_intermediate_statistics_from_deltas(deltas), audio_ids


def _worker(
        audio_ids, scores, single_label_ground_truths, audio_durations,
        segment_length=1., event_classes=None, time_decimals=6,
        output_queue=None,
):
    deltas = {}
    for audio_id in audio_ids:
        scores_k = scores[audio_id]
        timestamps, _ = validate_score_dataframe(
            scores_k, event_classes=event_classes)
        timestamps = np.round(timestamps, time_decimals)
        scores_k = scores_k[event_classes].to_numpy()
        if audio_durations is None:
            duration = max(
                [timestamps[-1]] + [
                    t_off for class_name in event_classes
                    for t_on, t_off, *_ in single_label_ground_truths[class_name][audio_id]
                ]
            )
        else:
            duration = audio_durations[audio_id]
        n_segments = int(np.ceil(duration / segment_length))
        segment_boundaries = np.round(
            np.arange(n_segments+1) * segment_length,
            time_decimals
        )
        segment_onsets = segment_boundaries[:-1]
        segment_offsets = segment_boundaries[1:]
        tp_deltas = {}
        for class_name in event_classes:
            gt = single_label_ground_truths[class_name][audio_id]
            if len(gt) == 0:
                tp_deltas[class_name] = np.zeros(n_segments, dtype=bool)
            else:
                tp_deltas[class_name] = np.any([
                    (segment_onsets < gt_offset)
                    * (segment_offsets > gt_onset)
                    * (segment_offsets > segment_onsets)
                    for gt_onset, gt_offset in
                    single_label_ground_truths[class_name][audio_id]
                ], axis=0)
        cp_scores = {class_name: [] for class_name in event_classes}
        for i in range(n_segments):
            idx_on = get_first_index_where(
                timestamps, "gt", segment_onsets[i]) - 1
            idx_on = max(idx_on, 0)
            idx_off = get_first_index_where(
                timestamps, "geq", segment_offsets[i])
            idx_off = min(idx_off, len(timestamps)-1)
            if idx_off <= idx_on:
                scores_ki = np.zeros(scores_k.shape[-1])
            else:
                scores_ki = np.max(scores_k[idx_on:idx_off], axis=0)
            for c, class_name in enumerate(event_classes):
                cp_scores[class_name].append(scores_ki[c])
        if audio_id not in deltas:
            deltas[audio_id] = {}
        for class_name in tp_deltas:
            deltas[audio_id][class_name] = (np.array(cp_scores[class_name]), {'tps': tp_deltas[class_name], 'fps': 1-tp_deltas[class_name]})
    if output_queue is not None:
        output_queue.put(deltas)
    return deltas



from sed_scores_eval.base_modules.roc import (
    roc_curve_from_intermediate_statistics,
    auroc_from_intermediate_statistics
)
from sed_scores_eval.clip_based.intermediate_statistics import accumulated_intermediate_statistics


def roc_curve(scores, ground_truth, *, deltas=None, num_jobs=1):
    """

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns: (all arrays sorted by corresponding false positive rate)
        true_positive_rates ((dict of) 1d np.ndarray): TPR values for all operating points
        false_positive_rates ((dict of) 1d np.ndarray): FPR values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding TPR-FPR pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        num_jobs=num_jobs,
    )
    return roc_curve_from_intermediate_statistics(intermediate_stats)


def auroc(scores, ground_truth, *, deltas=None, max_fpr=None, num_jobs=1):
    """compute area under ROC curve

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        max_fpr (float): maximum false positive rate up to which to compute partial auc
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        auroc ((dict of) float): mean and class-wise area under ROC curve
        roc_curve: ROC (TPR-FPR) curve(s) as provided by
            `roc_curve_from_intermediate_statistics`

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        num_jobs=num_jobs,
    )
    return auroc_from_intermediate_statistics(intermediate_stats, max_fpr=max_fpr)



from sed_scores_eval.base_modules.bootstrap import bootstrap_from_deltas
from sed_scores_eval.utils import parallel
from sed_scores_eval.base_modules.precision_recall import (
    single_fscore_from_intermediate_statistics,
    best_fscore_from_intermediate_statistics,
    precision_recall_curve_from_intermediate_statistics,
    fscore_curve_from_intermediate_statistics
)
from sed_scores_eval.collar_based.intermediate_statistics import (
    accumulated_intermediate_statistics, intermediate_statistics_deltas,
)


def precision_recall_curve(
        scores, ground_truth, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0.,
        time_decimals=6, num_jobs=1,
):
    """Compute collar-based precision-recall curve [1].

    [1] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns: (all arrays sorted by corresponding recall)
        precisions ((dict of) 1d np.ndarray): precision values for all operating points
        recalls ((dict of) 1d np.ndarray): recall values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding precision-recall pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return precision_recall_curve_from_intermediate_statistics(
        intermediate_stats
    )


def fscore_curve(
        scores, ground_truth, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0.,
        beta=1., time_decimals=6, num_jobs=1,
):
    """Compute collar-based f-scores with corresponding precisions, recalls and
    intermediate statistics for various operating points

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        beta: \beta parameter for f-score computation
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns: (all arrays sorted by corresponding score)
        f_beta ((dict of) 1d np.ndarray): f-score values  for all operating
            points
        precisions ((dict of) 1d np.ndarray): precision values for all operating points
        recalls ((dict of) 1d np.ndarray): recall values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding precision-recall pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps': 1d np.ndarray of true positive counts for each score
            'fps': 1d np.ndarray of false positive counts for each score
            'n_ref': integer number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return fscore_curve_from_intermediate_statistics(
        intermediate_stats, beta=beta,
    )


def fscore(
        scores, ground_truth, threshold, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0., beta=1.,
        return_onset_offset_dist_sum=False, time_decimals=6, num_jobs=1,
):
    """Get collar-based f-score with corresponding precision, recall and
    intermediate statistics for a specific decision threshold

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        threshold ((dict of) float): threshold that is to be evaluated.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        beta: \beta parameter for f-score computation
        return_onset_offset_dist_sum: If True, return summed distances between
            predicted and true on-/offsets (for true positive predictions),
            which allows to compute and compensate biases.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        fscore ((dict of) float): fscore value for threshold
        precision ((dict of) float): precision value for threshold
        recall ((dict of) float): recall value for threshold
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count for threshold
            'fps' (int): false positive count for threshold
            'n_ref' (int): number of ground truth events

    """
    if isinstance(scores, (list, tuple)) or isinstance(deltas, (list, tuple)):
        # batch input
        batch_size = [len(v) for v in [scores, deltas] if v is not None][0]
        f, p, r, stats = list(zip(*parallel.map(
            (scores, deltas), arg_keys=('scores', 'deltas'),
            func=fscore, max_jobs=num_jobs,
            ground_truth=ground_truth, threshold=threshold,
            onset_collar=onset_collar, offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate, beta=beta,
            return_onset_offset_dist_sum=return_onset_offset_dist_sum,
            time_decimals=time_decimals,
            num_jobs=num_jobs//batch_size,
        )))
        return f, p, r, stats
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        return_onset_offset_dist_sum=return_onset_offset_dist_sum,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return single_fscore_from_intermediate_statistics(
        intermediate_stats, threshold=threshold, beta=beta,
    )


def bootstrapped_fscore(
        scores, ground_truth, threshold, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0., beta=1.,
        return_onset_offset_dist_sum=False, time_decimals=6,
        n_folds=5, n_iterations=20, num_jobs=1,
):
    """

    Args:
        scores:
        ground_truth:
        threshold:
        deltas:
        onset_collar:
        offset_collar:
        offset_collar_rate:
        beta:
        return_onset_offset_dist_sum:
        time_decimals:
        n_folds:
        n_iterations:
        num_jobs:

    Returns:

    """
    if isinstance(scores, (list, tuple)) or isinstance(deltas, (list, tuple)):
        # batch input
        batch_size = [len(v) for v in [scores, deltas] if v is not None][0]
        f, p, r, stats = list(zip(*parallel.map(
            (scores, deltas), arg_keys=('scores', 'deltas'),
            func=bootstrapped_fscore, max_jobs=num_jobs,
            ground_truth=ground_truth, threshold=threshold,
            onset_collar=onset_collar, offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate, beta=beta,
            return_onset_offset_dist_sum=return_onset_offset_dist_sum,
            time_decimals=time_decimals,
            n_folds=n_folds, n_iterations=n_iterations,
            num_jobs=num_jobs//batch_size,
        )))
        return f, p, r, stats
    if deltas is None:
        deltas = intermediate_statistics_deltas(
            scores=scores, ground_truth=ground_truth,
            onset_collar=onset_collar, offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate,
            return_onset_offset_dist_sum=return_onset_offset_dist_sum,
            time_decimals=time_decimals, num_jobs=num_jobs,
        )
    return bootstrap_from_deltas(
        fscore, deltas,
        n_folds=n_folds, n_iterations=n_iterations, num_jobs=num_jobs,
        threshold=threshold, scores=None, ground_truth=ground_truth,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate, beta=beta,
        return_onset_offset_dist_sum=return_onset_offset_dist_sum,
        time_decimals=time_decimals,
    )


def best_fscore(
        scores, ground_truth, *, deltas=None,
        onset_collar, offset_collar, offset_collar_rate=0.,
        min_precision=0., min_recall=0., beta=1.,
        time_decimals=6, num_jobs=1,
):
    """Get the best possible (macro-averaged) collar-based f-score with
    corresponding precision, recall, intermediate statistics and decision
    threshold

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        min_precision: the minimum precision that must be achieved.
        min_recall: the minimum recall that must be achieved. If the
            constraint(s) cannot be achieved at any threshold, however,
            fscore, precision, recall and threshold of 0,1,0,inf are returned.
        beta: \beta parameter for f-score computation
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        f_beta ((dict of) float): best achievable f-score value
        precision ((dict of) float): precision value at best fscore
        recall ((dict of) float): recall value at best fscore
        threshold ((dict of) float): threshold to obtain best fscore which is
            centered between the score that the threshold has to fall below
            and the next smaller score which results in different intermediate
            statistics.
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count at best fscore
            'fps' (int): false positive count at best fscore
            'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return best_fscore_from_intermediate_statistics(
        intermediate_stats, beta=beta,
        min_precision=min_precision, min_recall=min_recall,
    )



import numpy as np
import multiprocessing
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from sed_scores_eval.base_modules import statistics
from sed_scores_eval.base_modules.io import parse_inputs


def intermediate_statistics_deltas(
        scores, ground_truth, *, num_jobs=1,
):
    if not isinstance(num_jobs, int) or num_jobs < 1:
        raise ValueError(
            f'num_jobs has to be an integer greater or equal to 1 but '
            f'{num_jobs} was given.'
        )
    scores, ground_truth, audio_ids = parse_inputs(
        scores, ground_truth, tagging=True)
    _, event_classes = validate_score_dataframe(scores[audio_ids[0]])

    if num_jobs == 1:
        deltas = _worker(audio_ids, scores, ground_truth, event_classes)
    else:
        deltas = {}
        queue = multiprocessing.Queue()
        shard_size = int(np.ceil(len(audio_ids) / num_jobs))
        shards = [
            audio_ids[i*shard_size:(i+1)*shard_size] for i in range(num_jobs)
            if i*shard_size < len(audio_ids)
        ]
        processes = [
            multiprocessing.Process(
                target=_worker,
                args=(
                    shard, scores, ground_truth, event_classes, queue
                ),
                daemon=True,
            )
            for shard in shards
        ]
        try:
            for p in processes:
                p.start()
            count = 0
            while count < len(shards):
                deltas_i = queue.get()
                assert len(deltas.keys() & deltas_i.keys()) == 0, sorted(deltas.keys() & deltas_i.keys())
                deltas.update(deltas_i)
                count += 1
        finally:
            for p in processes:
                p.terminate()
    return deltas


def accumulated_intermediate_statistics_from_deltas(deltas):
    multi_label_statistics = statistics.accumulated_intermediate_statistics_from_deltas(deltas)
    for _, stats_cls in multi_label_statistics.values():
        stats_cls['n_sys'] = stats_cls['tps'] + stats_cls['fps']
        stats_cls['n_ref'] = stats_cls['tps'][0]
        fns = stats_cls['n_ref'] - stats_cls['tps']
        stats_cls['tns'] = stats_cls['n_sys'][0] - stats_cls['n_sys'] - fns
    return multi_label_statistics


def accumulated_intermediate_statistics(scores, ground_truth, *, deltas=None, num_jobs=1):
    """

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            tags for each audio clip or a file path from where the ground truth
            can be loaded.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:

    """
    if deltas is None:
        scores, ground_truth, audio_ids = parse_inputs(scores, ground_truth, tagging=True)
        deltas = intermediate_statistics_deltas(
            scores=scores, ground_truth=ground_truth, num_jobs=num_jobs,
        )
    else:
        audio_ids = list(deltas.keys())
    return accumulated_intermediate_statistics_from_deltas(deltas), audio_ids


def _worker(audio_ids, scores, ground_truth, event_classes=None, output_queue=None):
    deltas = {}
    for audio_id in audio_ids:
        scores_k = scores[audio_id]
        timestamps, _ = validate_score_dataframe(
            scores_k, event_classes=event_classes)
        scores_k = scores_k[event_classes].to_numpy().max(0)
        gt_k = ground_truth[audio_id]
        if not all([
            class_name in event_classes for class_name in gt_k
        ]):
            unknown_tags = [
                class_name for class_name in gt_k
                if class_name not in event_classes
            ]
            raise ValueError(
                f'ground truth contains unknown tags. Unknown tags: '
                f'{unknown_tags}; Known tags: {event_classes};'
            )
        deltas[audio_id] = {}
        for c, class_name in enumerate(event_classes):
            deltas[audio_id][class_name] = ([scores_k[c]], {'tps': [class_name in gt_k], 'fps': [class_name not in gt_k]})
    if output_queue is not None:
        output_queue.put(deltas)
    return deltas



import numpy as np
from sed_eval.util import bipartite_match
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from sed_scores_eval.base_modules.ground_truth import event_counts_and_durations
from sed_scores_eval.base_modules import statistics
from sed_scores_eval.base_modules.io import parse_inputs, parse_ground_truth
from sed_scores_eval.base_modules.detection import onset_deltas


def intermediate_statistics_deltas(
    scores, ground_truth, onset_collar, offset_collar, offset_collar_rate=0.,
    return_onset_offset_dist_sum=False, time_decimals=6, num_jobs=1,
):
    """Compute collar-based intermediate statistics deltas for all audio files
    for all event classes and decision thresholds. See [1] for details about
    collar-based (event-based) evaluation. See [2] for details about the deltas
    and the joint computation of intermediate statistics for arbitrary decision
    thresholds.

    [1] Annamaria Mesaros, Toni Heittola, and Tuomas Virtanen,
    "Metrics for polyphonic sound event detection",
    Applied Sciences, vol. 6, pp. 162, 2016

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        return_onset_offset_dist_sum: If True, return summed distances between
            predicted and true on-/offsets (for true positive predictions),
            which allows to compute and compensate biases.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        deltas (dict of dicts of tuples): For each audio clip for each event class:
            change_point_scores (dict of dicts): 1d array of scores at which the
                intermediate statistics change, when the threshold falls below
                it, for that particular clip and class.
            dict of delta values: provides for each intermediate statistic a
                1d array of the delta (change), when the threshold falls below
                each of the change_point_scores.
                Required intermediate statistics deltas:
                  "tps": true positives count array
                  "fps": false positives count array

    """
    scores, ground_truth, keys = parse_inputs(scores, ground_truth)
    _, event_classes = validate_score_dataframe(
        scores[keys[0]])
    return statistics.intermediate_statistics_deltas(
        scores, ground_truth,
        intermediate_statistics_fn=statistics_fn,
        acceleration_fn=acceleration_fn,
        onset_collar=onset_collar, offset_collar=offset_collar,
        offset_collar_rate=offset_collar_rate,
        return_onset_offset_dist_sum=return_onset_offset_dist_sum,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )


def accumulated_intermediate_statistics_from_deltas(deltas, ground_truth):
    """Compute collar-based intermediate statistics over all audio files
    for all event classes and decision thresholds. See [1] for details about
    collar-based (event-based) evaluation. See [2] for details about the joint
    computation of intermediate statistics for arbitrary decision thresholds.

    [1] Annamaria Mesaros, Toni Heittola, and Tuomas Virtanen,
    "Metrics for polyphonic sound event detection",
    Applied Sciences, vol. 6, pp. 162, 2016

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        deltas (dict of dicts of tuples): For each audio clip for each event class:
            change_point_scores (dict of dicts): 1d array of scores at which the
                intermediate statistics change, when the threshold falls below
                it, for that particular clip and class.
            dict of delta values: provides for each intermediate statistic a
                1d array of the delta (change), when the threshold falls below
                each of the change_point_scores.
                Required intermediate statistics deltas:
                  "tps": true positives count array
                  "fps": false positives count array
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.

    Returns (dict of tuples): for each event class a tuple of 1d scores array
        and a dict of intermediate statistics with the following keys
        (where each array has the same length as the scores array):
            "tps": true positives count array
            "fps": false positives count array
            "n_ref": integer number of ground truth events

    """
    multi_label_statistics = statistics.accumulated_intermediate_statistics_from_deltas(deltas)
    n_ref, _ = event_counts_and_durations(
        ground_truth, event_classes=multi_label_statistics.keys()
    )
    return {
        class_name: (cp_scores_cls, {**stats_cls, 'n_ref': n_ref[class_name]})
        for class_name, (cp_scores_cls, stats_cls) in multi_label_statistics.items()
    }


def accumulated_intermediate_statistics(
    scores, ground_truth, *, deltas=None,
    onset_collar, offset_collar, offset_collar_rate=0.,
    return_onset_offset_dist_sum=False, time_decimals=6, num_jobs=1,
):
    """Cascade of `intermediate_statistics_deltas` and `accumulated_intermediate_statistics_from_deltas`.

    Args:
        scores:
        ground_truth:
        deltas:
        onset_collar:
        offset_collar:
        offset_collar_rate:
        return_onset_offset_dist_sum:
        time_decimals:
        num_jobs:

    Returns:

    """
    if deltas is None:
        scores, ground_truth, audio_ids = parse_inputs(scores, ground_truth)
        deltas = intermediate_statistics_deltas(
            scores, ground_truth,
            onset_collar=onset_collar, offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate,
            return_onset_offset_dist_sum=return_onset_offset_dist_sum,
            time_decimals=time_decimals, num_jobs=num_jobs,
        )
    else:
        audio_ids = list(deltas.keys())
        ground_truth = parse_ground_truth(
            ground_truth, audio_ids=audio_ids, additional_ids_ok=True)
    return accumulated_intermediate_statistics_from_deltas(deltas, ground_truth), audio_ids


def statistics_fn(
    detection_onset_times, detection_offset_times,
    target_onset_times, target_offset_times,
    other_onset_times, other_offset_times,
    onset_collar, offset_collar, offset_collar_rate=0.,
    return_onset_offset_dist_sum=False,
    time_decimals=6,
):
    """Compute collar-based intermediate statistics for a single audio and
    single target class given detected onset/offset times and target class
    ground truth onset/offset times

    Args:
        detection_onset_times (np.ndarray): (B, M) onset times of detected
            target class events with M being the number of detected target
            class events, and B being an independent dimension.
        detection_offset_times (np.ndarray): (B, M) offset times of detected
            target class events with M being the number of detected target
            class events, and B being an independent dimension. Note that it
            may include offset times which are equal to the corresponding onset
            time, which indicates that the event is inactive at that specific
            position b along the independent axis and must not be counted as a
            detection.
        target_onset_times (1d np.ndarray): onset times of target class ground
            truth events.
        target_offset_times (1d np.ndarray): offset times of target class
            ground truth events.
        other_onset_times (list of 1d np.ndarrays): onset times of other class
            ground truth events
        other_offset_times (list of 1d np.ndarrays): offset times of other
            class ground truth events
        onset_collar (float): allowed onset deviation in seconds
        offset_collar (float): (at least) allowed offset deviation in seconds
        offset_collar_rate (float): (at least) allowed offset deviation as a
            ratio of the length of the ground truth event, with the actual
            allowed offset deviation being:
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        return_onset_offset_dist_sum: If True, return summed distances between
            predicted and true on-/offsets (for true positive predictions),
            which allows to compute and compensate biases.
            #ToDo: add on-/offset bias compensation
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, detections with an onset or offset right on the
            boundary of the collar may be falsely counted as false detection
            because of small deviations due to limited floating point precision.

    Returns (dict of 1d np.ndarrays): dict of intermediate statistics with the
        following keys (where each array has the length B):
            "tps": true positives count array
            "fps": false positives count array

    """

    det_crit = detection_offset_times > detection_onset_times
    num_detections = det_crit.sum(-1)

    onset_dist = detection_onset_times[..., None] - target_onset_times
    onset_crit = np.round(
        np.abs(onset_dist) - onset_collar, decimals=time_decimals) <= 0.

    offset_collars = np.maximum(
        offset_collar,
        offset_collar_rate * (target_offset_times-target_onset_times),
    )
    offset_dist = detection_offset_times[..., None] - target_offset_times
    offset_crit = np.round(
        np.abs(offset_dist) - offset_collars, decimals=time_decimals) <= 0.
    hit_mat = det_crit[..., None] * onset_crit * offset_crit
    assert np.logical_or(hit_mat == 0, hit_mat == 1).all(), np.unique(hit_mat.flatten())
    invalid_detections = np.logical_or(
        np.any(hit_mat.sum(1) > 1, axis=1),
        np.any(hit_mat.sum(2) > 1, axis=1),
    )
    for idx in np.argwhere(invalid_detections).flatten():
        G = {}
        for det_idx, gt_idx in np.argwhere(hit_mat[idx]):
            if det_idx not in G:
                G[det_idx] = []
            G[det_idx].append(gt_idx)

        matching = sorted(bipartite_match(G).items())
        hit_mat[idx] = np.zeros_like(hit_mat[idx])
        for gt_idx, det_idx in matching:
            hit_mat[idx][det_idx, gt_idx] = 1
    tps = hit_mat.sum((1, 2))
    fps = num_detections - tps
    stats = {
        'tps': tps,
        'fps': fps,
    }
    if return_onset_offset_dist_sum:
        onset_dist = (hit_mat * onset_dist).sum((1, 2))
        offset_dist = (hit_mat * offset_dist).sum((1, 2))
        stats.update({
            'onset_dist_sum': onset_dist,
            'offset_dist_sum': offset_dist,
        })
    return stats


def acceleration_fn(
        scores, timestamps,
        target_onset_times, target_offset_times,
        other_onset_times, other_offset_times,
        onset_collar, offset_collar, offset_collar_rate=0.,
        return_onset_offset_dist_sum=False,
        time_decimals=6,
):
    # return np.unique(scores), None, None
    onset_deltas_ = onset_deltas(scores)
    change_points = np.abs(onset_deltas_) > .5
    if len(target_onset_times) == 0:
        cp_scores = scores[change_points]
        deltas = {
            'fps': onset_deltas_[change_points],
            'tps': np.zeros_like(cp_scores),
        }
        if return_onset_offset_dist_sum:
            deltas.update({
                'onset_dist_sum': np.zeros_like(cp_scores),
                'offset_dist_sum': np.zeros_like(cp_scores),
            })
        return None, cp_scores, deltas
    return None, None, None



import numpy as np
from functools import partial
from sed_scores_eval.utils import parallel
from sed_scores_eval.base_modules.bootstrap import bootstrap_from_deltas
from sed_scores_eval.base_modules.postprocessing import medfilt
from sed_scores_eval.intersection_based.intermediate_statistics import intermediate_statistics_deltas
from sed_scores_eval.intersection_based.psds import _unique_cummax_sort, psd_roc, psds_from_psd_roc, multi_class_psd_roc_from_single_class_psd_rocs


def psd_roc_postprocessing(
        scores, ground_truth, audio_durations, *, postprocessing_fn,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, num_jobs=1,
):
    """Compute PSD-ROC after applying post-processing.

    Args:
        scores:
        ground_truth:
        audio_durations:
        postprocessing_fn:
        dtc_threshold:
        gtc_threshold:
        cttc_threshold:
        alpha_ct:
        alpha_st:
        unit_of_time:
        max_efpr:
        time_decimals:
        num_jobs:

    Returns:

    """
    scores = postprocessing_fn(scores, lazy=True)
    return psd_roc(
        scores=scores, ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold,
        alpha_ct=alpha_ct, alpha_st=alpha_st,
        unit_of_time=unit_of_time, max_efpr=max_efpr,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )


def deltas_postprocessing(
        scores, ground_truth, postprocessing_fn, *,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        time_decimals=6, num_jobs=1,
):
    """Compute deltas after applying post-processing

    Args:
        scores:
        ground_truth:
        postprocessing_fn:
        dtc_threshold:
        gtc_threshold:
        cttc_threshold:
        time_decimals:
        num_jobs:

    Returns:

    """
    scores = postprocessing_fn(scores, lazy=True)
    return intermediate_statistics_deltas(
        scores=scores, ground_truth=ground_truth,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )


def merge_individual_rocs_into_overall_roc(rocs):
    """Combine operating points from multiple ROCS into a single RPC

    Args:
        rocs:

    Returns:

    """
    tprs, efprs, scores, filter_lengths = [
        np.concatenate(values) for values in list(zip(*rocs))
    ]
    return _unique_cummax_sort(tprs, efprs, scores, filter_lengths)


def postprocessing_independent_psds(
        scores, ground_truth, audio_durations, *, postprocessing_functions,
        scores_processed=None, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, num_jobs=1,
):
    """
    Computes post-processing independent Polyphonic Sound Detection Score (piPSDS) [3],
    which is the normalized area under pi-PSD-ROC. Similar to PSD-ROC [1, 2] the
    pi-PSD-ROC is computed as the mean of single class pi-ROCs plus a penalty on
    across class std. Single class pi-ROCs combine operating points from a set
    of different post-processings, such that for each eFPR the pi-TPR(eFPR) is
    given as the max TPR over the set of post-processings.

    [1] C. Bilen, G. Ferroni, F. Tuveri, J. Azcarreta and S. Krstulovic,
    "A Framework for the Robust Evaluation of Sound Event Detection",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2020, pp. 61–65

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    [3] J.Ebbers, R.Haeb-Umbach, and R.Serizel
    "Post-Processing Independent Evaluation of Sound Event Detection Systems",
    submitted to Detection and Classification of Acoustic Scenes and Events (DCASE) Workshop,
    2023

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        postprocessing_functions:
        scores_processed:
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        alpha_ct (float): parameter for penalizing cross triggers.
            More specifically, it is the weight of the cross trigger rate
            (averaged over all other classes) that is added to the False
            Positive Rate (FPR) yielding the effective FPR (eFPR). Default is 0.
        alpha_st (float): parameter for penalizing instability across classes.
            More specifically, it is the weight of the standard deviation of
            the per-class ROCs, that is subtracted from the mean of the
            per-class ROCs. Default is 0.
        unit_of_time (str): the unit of time \in {second, minute, hour} to be
            used for computation of the eFPR (which is defined as rate per unit
            of time). Default is hour.
        max_efpr (float): the maximum eFPR for which the system is evaluated,
            i.e., until where the Area under PSD ROC Curve is computed.
            Default is 100.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, e.g., a detection with an ground truth intersection
            exactly matching the DTC, may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        psds (float): Polyphonic Sound Detection Score (PSDS), i.e., the area
            under the PSD ROC Curve up to max_efpr normalized by max_efpr.
        psd_roc (tuple of 1d np.ndarrays): tuple of effective True Positive
            Rates and effective False Positive Rates.
        single_class_psd_rocs (dict of tuples of 1d np.ndarrays):
            tuple of True Positive Rates and effective False Positive Rates
            for each event class.

    """
    if (
        isinstance(scores, (list, tuple))
        or (isinstance(scores_processed, (list, tuple)) and isinstance(scores_processed[0], (list, tuple)))
        or (isinstance(deltas, (list, tuple)) and isinstance(deltas[0], (list, tuple)))
    ):
        batch_size = [len(v) for v in [scores, scores_processed, deltas] if v is not None][0]
        return list(zip(*parallel.map(
            (scores, scores_processed, deltas), ('scores', 'scores_processed', 'deltas'),
            func=postprocessing_independent_psds, max_jobs=num_jobs,
            postprocessing_functions=postprocessing_functions,
            ground_truth=ground_truth, audio_durations=audio_durations,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold, alpha_ct=alpha_ct,
            alpha_st=alpha_st, unit_of_time=unit_of_time, max_efpr=max_efpr,
            time_decimals=time_decimals,
            num_jobs=num_jobs//batch_size,
        )))
    (pi_effective_tp_rate, pi_effective_fp_rate), single_class_pi_psd_rocs, psd_rocs, single_class_psd_rocs = postprocessing_independent_psd_roc(
        scores=scores, ground_truth=ground_truth, audio_durations=audio_durations,
        postprocessing_functions=postprocessing_functions,
        scores_processed=scores_processed, deltas=deltas,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold, alpha_ct=alpha_ct, alpha_st=alpha_st,
        unit_of_time=unit_of_time, max_efpr=max_efpr,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    pi_psds_value = psds_from_psd_roc(
        pi_effective_tp_rate, pi_effective_fp_rate, max_efpr)
    single_class_pi_psds = {
        class_name: psds_from_psd_roc(tpr, efpr, max_efpr)
        for class_name, (tpr, efpr, *_) in single_class_pi_psd_rocs.items()
    }
    return (
        pi_psds_value,
        single_class_pi_psds,
        (pi_effective_tp_rate, pi_effective_fp_rate),
        single_class_pi_psd_rocs,
        psd_rocs,
        single_class_psd_rocs
    )


def postprocessing_independent_psd_roc(
        scores, ground_truth, audio_durations, *, postprocessing_functions,
        scores_processed=None, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, num_jobs=1,
):
    """Similar to PSD-ROC [1,2] the pi-PSD-ROC [3] is computed as the mean of
    single class pi-ROCs plus a penalty on across class std. Single class
    pi-ROCs combine operating points from a set of different post-processings,
    such that for each eFPR the pi-TPR(eFPR) is given as the max TPR over the
    set of post-processings.

    [1] C. Bilen, G. Ferroni, F. Tuveri, J. Azcarreta and S. Krstulovic,
    "A Framework for the Robust Evaluation of Sound Event Detection",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2020, pp. 61–65

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    [3] J.Ebbers, R.Haeb-Umbach, and R.Serizel
    "Post-Processing Independent Evaluation of Sound Event Detection Systems",
    submitted to Detection and Classification of Acoustic Scenes and Events (DCASE) Workshop,
    2023

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        postprocessing_functions:
        scores_processed:
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        alpha_ct (float): parameter for penalizing cross triggers.
            More specifically, it is the weight of the cross trigger rate
            (averaged over all other classes) that is added to the False
            Positive Rate (FPR) yielding the effective FPR (eFPR). Default is 0.
        alpha_st (float): parameter for penalizing instability across classes.
            More specifically, it is the weight of the standard deviation of
            the per-class ROCs, that is subtracted from the mean of the
            per-class ROCs. Default is 0.
        unit_of_time (str): the unit of time \in {second, minute, hour} to be
            used for computation of the eFPR (which is defined as rate per unit
            of time). Default is hour.
        max_efpr (float): the maximum eFPR for which the system is evaluated,
            i.e., until where the Area under PSD ROC Curve is computed.
            Default is 100.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, e.g., a detection with an ground truth intersection
            exactly matching the DTC, may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        etpr (1d np.ndarray): effective MFI True Positive Rates.
        efpr (1d np.ndarray): effective False Positive Rates.
        single_class_psd_rocs (dict of tuples of 1d np.ndarrays):
            tuple of MFI True Positive Rates and effective False Positive Rates
            for each event class.

    """
    if (
        isinstance(scores, (list, tuple))
        or (isinstance(scores_processed, (list, tuple)) and isinstance(scores_processed[0], (list, tuple)))
        or (isinstance(deltas, (list, tuple)) and isinstance(deltas[0], (list, tuple)))
    ):
        # batch input
        batch_size = [len(v) for v in [scores, scores_processed, deltas] if v is not None][0]
        return list(zip(*parallel.map(
            (scores, scores_processed, deltas), ('scores', 'scores_processed', 'deltas'),
            func=postprocessing_independent_psd_roc, max_jobs=num_jobs,
            postprocessing_functions=postprocessing_functions,
            ground_truth=ground_truth, audio_durations=audio_durations,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold, alpha_ct=alpha_ct,
            alpha_st=alpha_st, unit_of_time=unit_of_time, max_efpr=max_efpr,
            time_decimals=time_decimals,
            num_jobs=num_jobs//batch_size,
        )))
    if deltas is not None:
        assert isinstance(deltas, (list, tuple)), type(deltas)
        if scores_processed is None:
            scores_processed = len(deltas) * [None]
        if postprocessing_functions is not None:
            assert len(deltas) == len(postprocessing_functions), (len(deltas), len(postprocessing_functions))
    if scores_processed is not None:
        assert isinstance(scores_processed, (list, tuple)), type(scores_processed)
        if postprocessing_functions is not None:
            assert len(scores_processed) == len(postprocessing_functions), (len(scores_processed), len(postprocessing_functions))
        psd_rocs, single_class_psd_rocs = list(zip(*parallel.map(
            (scores_processed, deltas), arg_keys=('scores', 'deltas'),
            func=psd_roc, max_jobs=num_jobs,
            ground_truth=ground_truth, audio_durations=audio_durations,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold, alpha_ct=alpha_ct,
            alpha_st=alpha_st, unit_of_time=unit_of_time, max_efpr=max_efpr,
            time_decimals=time_decimals,
            num_jobs=num_jobs//len(scores_processed),
        )))
    else:
        assert isinstance(postprocessing_functions, (list, tuple)), type(postprocessing_functions)
        psd_rocs, single_class_psd_rocs = list(zip(*parallel.map(
            postprocessing_functions, arg_keys='postprocessing_fn',
            func=psd_roc_postprocessing, max_jobs=num_jobs,
            scores=scores, ground_truth=ground_truth, audio_durations=audio_durations,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold, alpha_ct=alpha_ct,
            alpha_st=alpha_st, unit_of_time=unit_of_time, max_efpr=max_efpr,
            time_decimals=time_decimals,
            num_jobs=1,
        )))

    single_class_pi_psd_rocs = {
        class_name: merge_individual_rocs_into_overall_roc([
            (
                *single_class_psd_rocs[m][class_name],
                np.full(
                    len(single_class_psd_rocs[m][class_name][-1]),
                    m,
                    dtype=int,
                )
            )
            for m in range(len(single_class_psd_rocs))
        ])
        for class_name in single_class_psd_rocs[0]
    }

    pi_psd_roc = multi_class_psd_roc_from_single_class_psd_rocs(
        single_class_pi_psd_rocs, alpha_st=alpha_st, max_efpr=max_efpr
    )
    return pi_psd_roc, single_class_pi_psd_rocs, psd_rocs, single_class_psd_rocs


def bootstrapped_postprocessing_independent_psds(
        scores, ground_truth, audio_durations, *, postprocessing_functions,
        scores_processed=None, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, n_folds=5, n_iterations=4, num_jobs=1,
):
    """

    Args:
        scores:
        ground_truth:
        audio_durations:
        postprocessing_functions:
        scores_processed:
        deltas:
        dtc_threshold:
        gtc_threshold:
        cttc_threshold:
        alpha_ct:
        alpha_st:
        unit_of_time:
        max_efpr:
        time_decimals:
        n_folds:
        n_iterations:
        num_jobs:

    Returns:

    """
    if (
        isinstance(scores, (list, tuple))
        or (isinstance(scores_processed, (list, tuple)) and isinstance(scores_processed[0], (list, tuple)))
        or (isinstance(deltas, (list, tuple)) and isinstance(deltas[0], (list, tuple)))
    ):
        # batch input
        batch_size = [len(v) for v in [scores, scores_processed, deltas] if v is not None][0]
        return list(zip(*parallel.map(
            (scores, scores_processed, deltas),
            arg_keys=('scores', 'scores_processed', 'deltas'),
            func=bootstrapped_postprocessing_independent_psds,
            postprocessing_functions=postprocessing_functions,
            max_jobs=num_jobs, ground_truth=ground_truth,
            audio_durations=audio_durations,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold, alpha_ct=alpha_ct,
            alpha_st=alpha_st, unit_of_time=unit_of_time, max_efpr=max_efpr,
            time_decimals=time_decimals,
            n_folds=n_folds, n_iterations=n_iterations,
            num_jobs=num_jobs//batch_size,
        )))
    if deltas is None:
        assert isinstance(postprocessing_functions, (list, tuple)), type(postprocessing_functions)
        if scores_processed is None:
            deltas = list(parallel.map(
                postprocessing_functions, arg_keys='postprocessing_fn',
                func=deltas_postprocessing, max_jobs=num_jobs,
                scores=scores, ground_truth=ground_truth,
                dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
                cttc_threshold=cttc_threshold, time_decimals=time_decimals,
                num_jobs=1,
            ))
        else:
            deltas = list(parallel.map(
                scores_processed, arg_keys='scores',
                func=intermediate_statistics_deltas, max_jobs=num_jobs,
                ground_truth=ground_truth,
                dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
                cttc_threshold=cttc_threshold, time_decimals=time_decimals,
                num_jobs=1,
            ))
    return bootstrap_from_deltas(
        postprocessing_independent_psds, deltas,
        n_folds=n_folds, n_iterations=n_iterations, num_jobs=num_jobs,
        scores=None, ground_truth=ground_truth, audio_durations=audio_durations,
        postprocessing_functions=postprocessing_functions,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold, alpha_ct=alpha_ct, alpha_st=alpha_st,
        unit_of_time=unit_of_time, max_efpr=max_efpr
    )


def median_filter_independent_psds(
        scores, ground_truth, audio_durations, *, median_filter_lengths_in_sec,
        scores_processed=None, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, num_jobs=1,
):
    postprocessing_functions = [
        partial(medfilt, filter_length_in_sec=filter_length_in_sec)
        for filter_length_in_sec in median_filter_lengths_in_sec
    ]
    return postprocessing_independent_psds(
        scores, ground_truth, audio_durations,
        postprocessing_functions=postprocessing_functions,
        scores_processed=scores_processed, deltas=deltas,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold, alpha_ct=alpha_ct, alpha_st=alpha_st,
        unit_of_time=unit_of_time, max_efpr=max_efpr,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )


def bootstrapped_median_filter_independent_psds(
        scores, ground_truth, audio_durations, *, median_filter_lengths_in_sec,
        scores_processed=None, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, n_folds=5, n_iterations=4, num_jobs=1,
):
    postprocessing_functions = [
        partial(medfilt, filter_length_in_sec=filter_length_in_sec)
        for filter_length_in_sec in median_filter_lengths_in_sec
    ]
    return bootstrapped_postprocessing_independent_psds(
        scores, ground_truth, audio_durations,
        postprocessing_functions=postprocessing_functions,
        scores_processed=scores_processed, deltas=deltas,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold, alpha_ct=alpha_ct, alpha_st=alpha_st,
        unit_of_time=unit_of_time, max_efpr=max_efpr,
        time_decimals=time_decimals, num_jobs=num_jobs,
        n_folds=n_folds, n_iterations=n_iterations,
    )



import numpy as np
import pandas as pd
from pathlib import Path
from sed_scores_eval.utils.array_ops import get_first_index_where
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from sed_scores_eval.base_modules.detection import scores_to_event_list
from sed_scores_eval.base_modules.io import (
    parse_inputs, write_detections_for_multiple_thresholds
)
from sed_scores_eval.intersection_based.psds import psds_from_psd_roc


def approximate_psds(
        scores, ground_truth, audio_durations,
        thresholds=np.linspace(.01, .99, 50), *,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        score_transform=None,
):
    """Reference psds implementation using the psds_eval package
    (https://github.com/audioanalytic/psds_eval), which, however, only
    approximates the PSD-ROC using a limited set of thresholds/operating points.
    This function is primarily used for testing purposes.

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        thresholds: the set of thresholds used to approximate the PSD-ROC.
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        alpha_ct (float): parameter for penalizing cross triggers.
            More specifically, it is the weight of the cross trigger rate
            (averaged over all other classes) that is added to the False
            Positive Rate (FPR) yielding the effective FPR (eFPR). Default is 0.
        alpha_st (float): parameter for penalizing instability across classes.
            More specifically, it is the weight of the standard deviation of
            the per-class ROCs, that is subtracted from the mean of the
            per-class ROCs. Default is 0.
        unit_of_time (str): the unit of time \in {second, minute, hour} to be
            used for computation of the eFPR (which is defined as rate per unit
            of time). Default is hour.
        max_efpr (float): the maximum eFPR for which the system is evaluated,
            i.e., until where the Area under PSD ROC Curve is computed.
            Default is 100.
        score_transform: a (non-linear) score transformation may be used before
            thresholding to obtain a better PSD-ROC approximation [1].
            [1] J.Ebbers, R.Serizel, and R.Haeb-Umbach
            "Threshold-Independent Evaluation of Sound Event Detection Scores",
            in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
            2022

    Returns:
        psds (float): Polyphonic Sound Detection Score (PSDS), i.e., the area
            under the approximated PSD ROC Curve up to max_efpr normalized
            by max_efpr.
        psd_roc (tuple of 1d np.ndarrays): tuple of effective True Positive
            Rates and effective False Positive Rates.
        single_class_psd_rocs (dict of tuples of 1d np.ndarrays):
            tuple of True Positive Rates and effective False Positive Rates
            for each event class.

    """
    import tempfile
    scores, _, keys = parse_inputs(scores, ground_truth)
    ground_truth = _parse_events(ground_truth)
    audio_format = ground_truth['filename'][0].split('.')[-1]
    validate_score_dataframe(scores[keys[0]])
    with tempfile.TemporaryDirectory() as tmp_dir:
        dir_path = Path(tmp_dir)
        write_detections_for_multiple_thresholds(
            scores, thresholds, dir_path, audio_format=audio_format,
            score_transform=score_transform, threshold_decimals=16,
        )
        return approximate_psds_from_detections_dir(
            dir_path, ground_truth, audio_durations, thresholds,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold,
            alpha_ct=alpha_ct, alpha_st=alpha_st,
            unit_of_time=unit_of_time, max_efpr=max_efpr,
            threshold_decimals=16,
        )


def approximate_psds_from_detections_dir(
        dir_path, ground_truth, audio_durations,
        thresholds=np.linspace(.01, .99, 50), *,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        threshold_decimals=3,
):
    """Reference psds implementation using the psds_eval package
    (https://github.com/audioanalytic/psds_eval), which, however, only
    approximates the PSD-ROC using a limited set of thresholds/operating points.
    This function is primarily used for testing purposes.

    Args:
        dir_path (str, pathlib.Path): path to directory where the detections
            for each threshold are saved.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        thresholds: the set of thresholds used to approximate the PSD-ROC.
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        alpha_ct (float): parameter for penalizing cross triggers.
            More specifically, it is the weight of the cross trigger rate
            (averaged over all other classes) that is added to the False
            Positive Rate (FPR) yielding the effective FPR (eFPR). Default is 0.
        alpha_st (float): parameter for penalizing instability across classes.
            More specifically, it is the weight of the standard deviation of
            the per-class ROCs, that is subtracted from the mean of the
            per-class ROCs. Default is 0.
        unit_of_time (str): the unit of time \in {second, minute, hour} to be
            used for computation of the eFPR (which is defined as rate per unit
            of time). Default is hour.
        max_efpr (float): the maximum eFPR for which the system is evaluated,
            i.e., until where the Area under PSD ROC Curve is computed.
            Default is 100.
        score_transform: a (non-linear) score transformation may be used before
            thresholding to obtain a better PSD-ROC approximation [1].
            [1] J.Ebbers, R.Serizel, and R.Haeb-Umbach
            "Threshold-Independent Evaluation of Sound Event Detection Scores",
            in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
            2022
        threshold_decimals:

    Returns:
        psds (float): Polyphonic Sound Detection Score (PSDS), i.e., the area
            under the approximated PSD ROC Curve up to max_efpr normalized
            by max_efpr.
        psd_roc (tuple of 1d np.ndarrays): tuple of effective True Positive
            Rates and effective False Positive Rates.
        single_class_psd_rocs (dict of tuples of 1d np.ndarrays):
            tuple of True Positive Rates and effective False Positive Rates
            for each event class.

    """
    assert np.all(np.abs(thresholds - np.round(thresholds, threshold_decimals)) < 1e-15), (threshold_decimals, thresholds)
    assert np.all(thresholds == np.unique(thresholds)), thresholds
    from psds_eval import PSDSEval
    ground_truth = _parse_events(ground_truth)
    audio_durations = _parse_audio_durations(audio_durations)
    dir_path = Path(dir_path)
    psds_eval = PSDSEval(
        dtc_threshold=dtc_threshold,
        gtc_threshold=gtc_threshold,
        cttc_threshold=1. if cttc_threshold is None else cttc_threshold,
        ground_truth=ground_truth,
        metadata=audio_durations,
        duration_unit=unit_of_time,
    )
    psds_eval.clear_all_operating_points()

    for i, threshold in enumerate(thresholds):
        tsv = dir_path / '{:.Xf}.tsv'.replace('X', str(threshold_decimals)).format(threshold)
        print(f"Adding operating point {i+1}/{len(thresholds)}", end="\r")
        det = pd.read_csv(tsv, sep="\t")
        info = {"name": f"Op {i+1:02d}", "threshold": threshold}
        psds_eval.add_operating_point(det, info=info)

    # compute the PSDS of the system represented by its operating points
    psds_ = psds_eval.psds(
        max_efpr=max_efpr,
        alpha_st=alpha_st,
        alpha_ct=alpha_ct,
    )
    efpr = psds_.plt.xp
    etpr = psds_.plt.yp
    _, _, tpr_vs_efpr = psds_eval.psd_roc_curves(alpha_ct=alpha_ct)

    single_class_psd_rocs = {
        psds_eval.class_names[i]: (tpr_vs_efpr.yp[i], tpr_vs_efpr.xp)
        for i in range(len(tpr_vs_efpr.yp))
    }

    if max_efpr is not None:
        cutoff_idx = get_first_index_where(efpr, "gt", max_efpr)
        etpr = np.array(etpr[:cutoff_idx].tolist() + [etpr[cutoff_idx-1]])
        efpr = np.array(efpr[:cutoff_idx].tolist() + [max_efpr])
        for key, roc in single_class_psd_rocs.items():
            cutoff_idx = get_first_index_where(roc[1], "gt", max_efpr)
            single_class_psd_rocs[key] = (
                roc[0][:cutoff_idx], roc[1][:cutoff_idx]
            )

    single_class_psds = {
        class_name: psds_from_psd_roc(tpr, efpr, max_efpr)
        for class_name, (tpr, efpr) in single_class_psd_rocs.items()
    }
    return psds_.value, single_class_psds, (etpr, efpr), single_class_psd_rocs


def fscore(scores, ground_truth, threshold, *, dtc_threshold, gtc_threshold, beta=1.):

    from psds_eval import PSDSEval

    scores, ground_truth, audio_ids = parse_inputs(scores, ground_truth)
    detections = _parse_events(scores_to_event_list(scores, threshold))
    gt = _parse_events(ground_truth)
    audio_durations = _parse_audio_durations({clip_id: scores_df['offset'].to_numpy()[-1] for clip_id, scores_df in scores.items()})
    psds = PSDSEval(
        ground_truth=gt,
        metadata=audio_durations,
        dtc_threshold=dtc_threshold,
        gtc_threshold=gtc_threshold,
    )
    if detections.empty:
        _, event_classes = validate_score_dataframe(scores[audio_ids[0]])
        f_macro = .0
        f = {event_class: .0 for event_class in event_classes}
    else:
        f_macro, f = psds.compute_macro_f_score(detections, beta=beta)
    for event_class in f:
        if np.isnan(f[event_class]):
            f[event_class] = 0.0
    f['macro_average'] = f_macro
    return f


def _parse_events(events):
    if isinstance(events, pd.DataFrame):
        return events
    elif isinstance(events, dict):
        events = [(f"{key}.wav", *event) for key in sorted(events.keys()) for event in events[key]]
        events = pd.DataFrame(
            events,
            columns=['filename', 'onset', 'offset', 'event_label']
        )
    else:
        assert isinstance(events, (str, Path)), type(events)
        events = pd.read_csv(events, sep="\t")
    return events


def _parse_audio_durations(audio_durations):
    if isinstance(audio_durations, pd.DataFrame):
        return audio_durations
    elif isinstance(audio_durations, dict):
        audio_durations = [(f"{key}.wav", audio_durations[key]) for key in audio_durations.keys()]
        audio_durations = pd.DataFrame(
            audio_durations, columns=['filename', 'duration']
        )
    else:
        assert isinstance(audio_durations, (str, Path)), type(audio_durations)
        audio_durations = pd.read_csv(audio_durations, sep="\t")
    return audio_durations



import builtins
from functools import partial
from concurrent.futures import ProcessPoolExecutor


def _map_function_wrapper(args, arg_keys, func, **kwargs):
    if isinstance(arg_keys, (list, tuple)):
        assert len(args) == len(arg_keys), (len(args), len(arg_keys))
        return func(**{k: v for k, v in zip(arg_keys, args)}, **kwargs)
    assert isinstance(arg_keys, str), type(arg_keys)
    return func(**{arg_keys: args}, **kwargs)


def map(args, arg_keys, func, max_jobs, **kwargs):
    map_fn = partial(_map_function_wrapper, arg_keys=arg_keys, func=func, **kwargs)

    if isinstance(arg_keys, (list, tuple)):
        assert len(args) == len(arg_keys), (len(args), len(arg_keys))
        lengths = {len(arg) for arg in args if arg is not None}
        assert len(lengths) == 1, "all non-None args must have the same length."
        length = list(lengths)[0]
        args = list(zip(*[length * [None] if arg is None else arg for arg in args]))
        num_jobs = min(max_jobs, length)
    else:
        num_jobs = min(max_jobs, len(args))

    if num_jobs > 1:
        with ProcessPoolExecutor(num_jobs) as ex:
            return list(ex.map(map_fn, args))
    else:
        return list(builtins.map(map_fn, args))



import numpy as np
from sed_scores_eval.utils.array_ops import get_first_index_where


def staircase_auc(y, x, max_x=None):
    """Area under Curve (AUC) with staircase type interpolation

    Args:
        y (1d np.ndarray): y values
        x (1d np.ndarray): x values
        max_x: maximal x value. If not None curve is interpolated up to this x
            value. Else AUC is computed up to the maximal value in x array.

    Returns:
        auc: AUC value

    >>> staircase_auc(np.array([0.,1.,2.,3.]), np.array([0.,1.,2.,3.]))
    3.0
    >>> staircase_auc(np.array([0.,1.,2.,3.]), np.array([0.,1.,1.1,1.2]))
    0.2999999999999998
    >>> staircase_auc(np.array([0.,1.,2.,3.]), np.array([0.,1.,2.,3.]), max_x=2.5)
    2.0
    >>> staircase_auc(np.array([0.,1.,2.,3.]), np.array([0.,1.,2.,3.]), max_x=10.)
    24.0
    """
    sort_idx = np.argsort(x)
    x = x[sort_idx]
    y = y[sort_idx]
    if max_x is None:
        y = y[:-1]
    else:
        cutoff_idx = get_first_index_where(x, "gt", max_x)
        x = np.concatenate((x[:cutoff_idx], [max_x]))
        y = y[:cutoff_idx]
    widths = x[1:] - x[:-1]
    areas = y * widths
    auc = sum(areas)
    return auc


def get_curve_idx_for_threshold(scores, threshold):
    """get that index of a curve that corresponds to a given threshold
    
    Args:
        scores: 
        threshold: 

    Returns:

    """
    if not np.isscalar(threshold):
        raise ValueError(
            f'threshold_of_interest must be scalar but {threshold}'
            f' was given'
        )
    sort_idx = np.argsort(scores)
    idx = get_first_index_where(scores[sort_idx], 'gt', threshold)
    if idx == len(sort_idx):
        assert scores[sort_idx[-1]] == np.inf, scores[sort_idx[-1]]
        idx = -1
    return sort_idx[idx]


def xsort(y, x, *args):
    sort_idx = sorted(np.arange(len(x)).tolist(), key=lambda i: (x[i], y[i]))

    def sort_arg(stat):
        if np.isscalar(stat):
            return stat
        if isinstance(stat, (list, tuple)):
            return [sort_arg(stat_i) for stat_i in stat]
        if isinstance(stat, dict):
            return {key: sort_arg(stat_i) for key, stat_i in stat.items()}
        return stat[sort_idx]

    return (
        y[sort_idx], x[sort_idx], *[sort_arg(arg) for arg in args],
    )



import numpy as np
import multiprocessing
from sed_scores_eval.utils.nested import flatten, deflatten
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from sed_scores_eval.base_modules.ground_truth import multi_label_to_single_label_ground_truths
from sed_scores_eval.base_modules.cy_detection import onset_offset_curves
from sed_scores_eval.base_modules.io import parse_inputs


def intermediate_statistics_deltas(
        scores, ground_truth, intermediate_statistics_fn, *,
        acceleration_fn=None, num_jobs=1,
        **intermediate_statistics_fn_kwargs
):
    """Core function of this package. It computes the deltas of intermediate
    statistics for single audio files as described in our paper [1] for all
    files in the dataset. Note that this function assumes intermediate
    statistics to be 0 for a decision threshold of infinity, i.e., when no event
    is detected. So the intermediate statistics have to be defined accordingly.

    [1] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        scores (dict, str, pathlib.Path): dict of multi-label SED score
            DataFrames (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        intermediate_statistics_fn (callable): a function returning a dict of
            intermediate statistics for a single target class and a single
            audio file by taking (at least) the following key word arguments
            (See collar_based.intermediate_statistics.statistics_fn or
            intersection_based.intermediate_statistics.statistics_fn for
            examples):
              detection_onset_times (np.ndarray): (B, M) onset times of
                detected target class events with M being the number of
                detected target class events, and B being an independent
                dimension.
              detection_offset_times (np.ndarray): (B, M) offset times of
                detected target class events with M being the number of
                detected target class events, and B being an independent
                dimension. Note that it may include offset times which are
                equal to the corresponding onset time, which indicates that the
                event is inactive at that specific position b along the
                independent axis and must not be counted as a detection.
              target_onset_times (1d np.ndarray): onset times of target class
                ground truth events.
              target_offset_times (1d np.ndarray): offset times of target class
                ground truth events.
              other_onset_times (list of 1d np.ndarrays): onset times of other
                class ground truth events
              other_offset_times (list of 1d np.ndarrays): offset times of
                other class ground truth events
        acceleration_fn (callable): a function returning a reduced set of
            change point candidates and/or directly the change point scores with
            corresponding intermediate statistic deltas.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.
        **intermediate_statistics_fn_kwargs: some other key word arguments for
            intermediate_statistics_fn, e.g., the collar in collar-based
            evaluation.

    Returns:
        deltas (dict of dicts of tuples): For each audio clip for each event class:
            change_point_scores (dict of dicts): 1d array of scores at which the
                intermediate statistics change, when the threshold falls below
                it, for that particular clip and class.
            dict of delta values: provides for each intermediate statistic a
                1d array of the delta (change), when the threshold falls below
                each of the change_point_scores.

    """
    if not isinstance(num_jobs, int) or num_jobs < 1:
        raise ValueError(
            f'num_jobs has to be an integer greater or equal to 1 but '
            f'{num_jobs} was given.'
        )
    scores, ground_truth, audio_ids = parse_inputs(scores, ground_truth)

    _, event_classes = validate_score_dataframe(scores[audio_ids[0]])
    single_label_ground_truths = multi_label_to_single_label_ground_truths(
        ground_truth, event_classes)

    if num_jobs == 1:
        deltas = _worker(
            audio_ids, scores, single_label_ground_truths,
            intermediate_statistics_fn, intermediate_statistics_fn_kwargs,
            acceleration_fn,
        )
    else:
        queue = multiprocessing.Queue()
        shard_size = int(np.ceil(len(audio_ids) / num_jobs))
        shards = [
            audio_ids[i*shard_size:(i+1)*shard_size] for i in range(num_jobs)
            if i*shard_size < len(audio_ids)
        ]
        processes = [
            multiprocessing.Process(
                target=_worker,
                args=(
                    shard, scores, single_label_ground_truths,
                    intermediate_statistics_fn,
                    intermediate_statistics_fn_kwargs,
                    acceleration_fn,
                    queue
                ),
                daemon=True,
            )
            for shard in shards
        ]
        try:
            for p in processes:
                p.start()
            deltas = {}
            count = 0
            while count < len(shards):
                deltas_i = queue.get()
                assert len(deltas.keys() & deltas_i.keys()) == 0, sorted(deltas.keys() & deltas_i.keys())
                deltas.update(deltas_i)
                count += 1
        finally:
            for p in processes:
                p.terminate()
    return deltas


def accumulated_intermediate_statistics_from_deltas(deltas):
    """
    Takes change_point_scores and deltas from `intermediate_statistics_deltas`.
    It then brings deltas from all audio clips in a single list sorted w.r.t.
    core values and computes intermediate statistics (over all clips) at various
    operating points by a cumulative sum over the deltas as described in our
    paper [1].

    Args:
        deltas (dict of dicts of tuples): For each audio clip for each event class:
            change_point_scores (dict of dicts): 1d array of scores at which the
                intermediate statistics change, when the threshold falls below
                it, for that particular clip and class.
            dict of delta values: provides for each intermediate statistic a
                1d array of the delta (change), when the threshold falls below
                each of the change_point_scores.

    Returns (dict of tuples): for each event class:
        - unique scores (1d np.ndarray) for which the intermediate statistics
            change when the threshold falls below it.
        - intermediate statistics (dict of 1d np.ndarray): dict of
            arrays of intermediate statistics for each of the scores.

    """
    audio_ids = list(deltas.keys())
    event_classes = list(deltas[audio_ids[0]].keys())
    acc_deltas = {
        class_name: ([], {key: [] for key in deltas[audio_ids[0]][class_name][1]})
        for class_name in event_classes
    }
    for audio_id in audio_ids:
        for class_name in event_classes:
            acc_deltas[class_name][0].append(deltas[audio_id][class_name][0])
            for key in acc_deltas[class_name][1]:
                acc_deltas[class_name][1][key].append(
                    deltas[audio_id][class_name][1][key])

    return {
        class_name: _intermediate_statistics_from_deltas(
            np.concatenate(acc_deltas[class_name][0]),
            {
                key: np.concatenate(acc_deltas[class_name][1][key])
                for key in acc_deltas[class_name][1]
            }
        )
        for class_name in event_classes
    }


def accumulated_intermediate_statistics(
        scores, ground_truth, intermediate_statistics_fn, *,
        acceleration_fn=None, num_jobs=1,
        **intermediate_statistics_fn_kwargs
):
    """Cascade of `intermediate_statistics_deltas` and `accumulated_intermediate_statistics_from_deltas`.

    Args:
        scores:
        ground_truth:
        intermediate_statistics_fn:
        acceleration_fn:
        num_jobs:
        **intermediate_statistics_fn_kwargs:

    Returns:

    """
    deltas = intermediate_statistics_deltas(
        scores, ground_truth, intermediate_statistics_fn,
        acceleration_fn=acceleration_fn, num_jobs=num_jobs,
        **intermediate_statistics_fn_kwargs
    )
    return accumulated_intermediate_statistics_from_deltas(deltas)


def _worker(
        audio_ids, scores, single_label_ground_truths,
        intermediate_statistics_fn, intermediate_statistics_fn_kwargs,
        acceleration_fn=None, output_queue=None
):
    deltas = {}
    _, event_classes = validate_score_dataframe(scores[audio_ids[0]])
    for audio_id in audio_ids:
        scores_for_key = scores[audio_id]
        timestamps, _ = validate_score_dataframe(
            scores_for_key, event_classes=event_classes)
        scores_for_key = scores_for_key[event_classes].to_numpy()
        deltas[audio_id] = {}
        gt_onset_times = {}
        gt_offset_times = {}
        for class_name in event_classes:
            gt = single_label_ground_truths[class_name][audio_id]
            if gt:
                current_onset_times, current_offset_times = np.array(gt).T
            else:
                current_onset_times = current_offset_times = np.empty(0)
            gt_onset_times[class_name] = current_onset_times
            gt_offset_times[class_name] = current_offset_times
        for c, class_name in enumerate(event_classes):
            target_onset_times = gt_onset_times[class_name]
            target_offset_times = gt_offset_times[class_name]
            other_onset_times = {ocls: onset_time for ocls, onset_time in gt_onset_times.items() if ocls != class_name}
            other_offset_times = {ocls: offset_time for ocls, offset_time in gt_offset_times.items() if ocls != class_name}
            if acceleration_fn is None:
                change_point_candidates = cp_scores_c = deltas_c = None
            else:
                change_point_candidates, cp_scores_c, deltas_c = acceleration_fn(
                    scores=scores_for_key[:, c], timestamps=timestamps,
                    target_onset_times=target_onset_times,
                    target_offset_times=target_offset_times,
                    other_onset_times=other_onset_times,
                    other_offset_times=other_offset_times,
                    **intermediate_statistics_fn_kwargs,
                )
                assert not (cp_scores_c is None) ^ (deltas_c is None)
            if cp_scores_c is None:
                (
                    unique_scores, detection_onset_times, detection_offset_times,
                ) = onset_offset_curves(
                    scores_for_key[:, c], timestamps, change_point_candidates
                )
                stats = intermediate_statistics_fn(
                    detection_onset_times=detection_onset_times,
                    detection_offset_times=detection_offset_times,
                    target_onset_times=target_onset_times,
                    target_offset_times=target_offset_times,
                    other_onset_times=other_onset_times,
                    other_offset_times=other_offset_times,
                    **intermediate_statistics_fn_kwargs,
                )
                cp_scores_c, deltas_c = _deltas_from_intermediate_statistics(
                    unique_scores, stats
                )
            if audio_id not in deltas:
                deltas[audio_id] = {}
            deltas[audio_id][class_name] = (cp_scores_c, deltas_c)
    if output_queue is not None:
        output_queue.put(deltas)
    return deltas


def _deltas_from_intermediate_statistics(scores, intermediate_stats):
    """compute deltas for intermediate statistics for single audio and single
    target class

    Args:
        scores (1d np.ndarray): single class SED scores from a single audio.
        intermediate_stats (dict of 1d np.ndarrays): dict of arrays of
            intermediate statistics for each of the scores

    Returns:
        change_point_scores (1d np.ndarray): array of scores for which the
            intermediate statistics change when the threshold falls below it.
        deltas (dict of 1d np.ndarrays): dict of arrays of the changes (deltas)
            in each intermediate statistic at each of the change point scores.

    """
    scores_unique, unique_idx = np.unique(scores, return_index=True)
    intermediate_stats = flatten(intermediate_stats, sep='\\')
    stat_keys = list(intermediate_stats.keys())
    stats = np.stack([
        intermediate_stats[key][unique_idx]
        for key in stat_keys
    ], axis=1)
    deltas = stats - np.concatenate((stats[1:], np.zeros_like(stats[:1])))

    # filter scores where nothing changes
    any_delta = np.abs(deltas).sum(1) > 0
    change_indices = np.argwhere(any_delta).flatten()
    change_point_scores = scores_unique[change_indices]
    deltas = {
        key: deltas_i[change_indices]
        for key, deltas_i in zip(stat_keys, deltas.T)
    }
    return change_point_scores, deltas


def _intermediate_statistics_from_deltas(change_point_scores, deltas):
    """sort and cumsum the deltas from all audio for each intermediate statistic

    Args:
        change_point_scores (1d np.ndarray): concatenation of single class SED scores from
            all audios.
        deltas (dict of 1d np.ndarrays): dict of concatenations of the changes
            (deltas) in each intermediate statistic at each of the scores.

    Returns:
        scores_unique:
        stats:
    """
    scores_unique, inverse_idx = np.unique(change_point_scores, return_inverse=True)
    b = len(scores_unique)
    scores_unique = np.concatenate((scores_unique, [np.inf]))
    stat_keys = list(deltas.keys())
    deltas = np.stack([deltas[key] for key in stat_keys], axis=1)
    deltas_unique = np.zeros((b, *deltas.shape[1:]))
    np.add.at(deltas_unique, inverse_idx, deltas)
    stats = np.concatenate((
        np.cumsum(deltas_unique[::-1], axis=0)[::-1],
        np.zeros_like(deltas_unique[:1])
    ))
    stats = deflatten({
        key: stats_i for key, stats_i in zip(stat_keys, stats.T)
    }, sep='\\')
    return scores_unique, stats
